#include <stdio.h>
#include "utils.h"

/**
 * ~TODO 3~
 * Modify the kernel below such as each element of the 
 * array will be now equal to 0 if it is an even number
 * or 1, if it is an odd number
 */
__global__ void kernel_parity_id(int *a, int N) {
    unsigned int i = threadIdx.x + blockDim.x * blockIdx.x;
    a[i] = a[i] % 2;
}

/**
 * ~TODO 4~
 * Modify the kernel below such as each element will
 * be equal to the BLOCK ID this computation takes
 * place.
 */
__global__ void kernel_block_id(int *a, int N) {
    unsigned int i = threadIdx.x + blockDim.x * blockIdx.x;
    a[i] = blockIdx.x;
}

/**
 * ~TODO 5~
 * Modify the kernel below such as each element will
 * be equal to the THREAD ID this computation takes
 * place.
 */ 
__global__ void kernel_thread_id(int *a, int N) {
    unsigned int i = threadIdx.x + blockDim.x * blockIdx.x;
    a[i] = threadIdx.x;
}

int main(void) {
    int nDevices;

    // Get the number of CUDA-capable GPU(s)
    cudaGetDeviceCount(&nDevices);

    /**
     * ~TODO 1~
     * For each device, show some details in the format below, 
     * then set as active device the first one (assuming there
     * is at least CUDA-capable device). Pay attention to the
     * type of the fields in the cudaDeviceProp structure.
     *
     * Device number: <i>
     *      Device name: <name>
     *      Total memory: <mem>
     *      Memory Clock Rate (KHz): <mcr>
     *      Memory Bus Width (bits): <mbw>
     * 
     * Hint: look for cudaGetDeviceProperties and cudaSetDevice in
     * the Cuda Toolkit Documentation. 
     */
    for (int i = 0; i < nDevices; ++i) {
        cudaDeviceProp prop;
        cudaGetDeviceProperties(&prop, i);

        printf("Device number: %d\n", i);
        printf("\tDevice name: %s\n", prop.name);
        printf("\tTotal memory: %lu\n", prop.totalGlobalMem);
        printf("\tMemory Clock Rate: %d\n", prop.memoryClockRate);
        printf("\tMemory Bus Width: %d\n", prop.memoryBusWidth);
    }

    cudaSetDevice(0);

    /**
     * ~TODO 2~
     * With information from example_2.cu, allocate an array with
     * integers (where a[i] = i). Then, modify the three kernels
     * above and execute them using 4 blocks, each with 4 threads.
     *
     * You can use the fill_array(int *a, int n) function (from utils)
     * to fill your array as many times you want.
     */ 
    int *device_array = 0;
    int *host_array = 0;
    int N = 1024;

    host_array = (int *) malloc(N * sizeof(int));
    cudaMalloc((void **) &device_array, N * sizeof(int));

    if (host_array == 0 || device_array == 0) {
        printf("[HOST] Couldn't allocate memory\n");
    	return 1;
    }

    fill_array_int(host_array, N); 
    cudaMemcpy(device_array, host_array, N * sizeof(int), cudaMemcpyHostToDevice);

    /**
     *  ~TODO 3~
     * Execute kernel_parity_id kernel and then copy from
     * the device to the host; call cudaDeviceSynchronize()
     * after a kernel execution for safety purposes.
     */ 
    kernel_parity_id<<<N / 4, 4>>>(device_array, N);
    cudaDeviceSynchronize();

    cudaMemcpy(host_array, device_array, N * sizeof(int), cudaMemcpyDeviceToHost);
    check_task_2(3, host_array);

    /**
     * ~TODO 4~
     * Execute kernel_block_id kernel and then copy from 
     * the device to the host;
     */
    kernel_block_id<<<N / 4, 4>>>(device_array, N);
    cudaDeviceSynchronize();

    cudaMemcpy(host_array, device_array, N * sizeof(int), cudaMemcpyDeviceToHost);
    check_task_2(4, host_array);

    /**
     * ~TODO 5~
     * Execute kernel_thread_id kernel and then copy from 
     * the device to the host;
     */
    kernel_thread_id<<<N / 4, 4>>>(device_array, N);
    cudaDeviceSynchronize();

    cudaMemcpy(host_array, device_array, N * sizeof(int), cudaMemcpyDeviceToHost);
    check_task_2(5, host_array);

    // Free the memory
    free(host_array);
    cudaFree(device_array);
    
    return 0;
}